@@ -37,6 +37,8 @@
 
 #include "gnome-rr-private.h"
 
+/* From xf86drmMode.h: it's ABI so it won't change */
+#define DRM_MODE_FLAG_INTERLACE			(1<<4)
 
 enum {
     SCREEN_PROP_0,
@@ -52,7 +54,7 @@ enum {
     SCREEN_SIGNAL_LAST,
 };
 
-gint screen_signals[SCREEN_SIGNAL_LAST];
+static gint screen_signals[SCREEN_SIGNAL_LAST];
 
 struct GnomeRROutput
 {
@@ -62,6 +64,7 @@ struct GnomeRROutput
     
     char *		name;
     char *		display_name;
+    char *		connector_type;
     GnomeRRCrtc *	current_crtc;
     GnomeRRCrtc **	possible_crtcs;
     GnomeRROutput **	clones;
@@ -113,6 +116,7 @@ struct GnomeRRMode
     int			height;
     int			freq;		/* in mHz */
     gboolean		tiled;
+    guint32             flags;
 };
 
 /* GnomeRRCrtc */
@@ -268,8 +272,8 @@ has_similar_mode (GnomeRROutput *output, GnomeRRMode *mode)
 {
     int i;
     GnomeRRMode **modes = gnome_rr_output_list_modes (output);
-    int width = gnome_rr_mode_get_width (mode);
-    int height = gnome_rr_mode_get_height (mode);
+    guint width = gnome_rr_mode_get_width (mode);
+    guint height = gnome_rr_mode_get_height (mode);
 
     for (i = 0; modes[i] != NULL; ++i)
     {
@@ -291,8 +295,8 @@ _gnome_rr_output_get_tiled_display_size (GnomeRROutput *output,
 					 int *total_width, int *total_height)
 {
     GnomeRRTile tile;
-    int ht, vt, i;
-    int total_h = 0, total_w = 0;
+    guint ht, vt;
+    int i, total_h = 0, total_w = 0;
 
     if (!_gnome_rr_output_get_tile_info (output, &tile))
 	return FALSE;
@@ -427,7 +431,7 @@ fill_screen_info_from_resources (ScreenInfo *info,
 				 int         max_width,
 				 int         max_height)
 {
-    int i;
+    guint i;
     GPtrArray *a;
     GnomeRRCrtc **crtc;
     GnomeRROutput **output;
@@ -451,13 +455,10 @@ fill_screen_info_from_resources (ScreenInfo *info,
     a = g_ptr_array_new ();
     for (i = 0; i < ncrtc; ++i)
     {
-	GnomeRRCrtc *crtc;
-
 	g_variant_get_child (crtcs, i, META_CRTC_STRUCT, &id,
 			     NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
-	crtc = crtc_new (info, id);
 
-	g_ptr_array_add (a, crtc);
+	g_ptr_array_add (a, crtc_new (info, id));
     }
     g_ptr_array_add (a, NULL);
     info->crtcs = (GnomeRRCrtc **)g_ptr_array_free (a, FALSE);
@@ -465,13 +466,10 @@ fill_screen_info_from_resources (ScreenInfo *info,
     a = g_ptr_array_new ();
     for (i = 0; i < noutput; ++i)
     {
-	GnomeRROutput *output;
- 
 	g_variant_get_child (outputs, i, META_OUTPUT_STRUCT, &id,
 			     NULL, NULL, NULL, NULL, NULL, NULL, NULL);
-	output = output_new (info, id);
 
-	g_ptr_array_add (a, output);
+	g_ptr_array_add (a, output_new (info, id));
     }
     g_ptr_array_add (a, NULL);
     info->outputs = (GnomeRROutput **)g_ptr_array_free (a, FALSE);
@@ -479,13 +477,10 @@ fill_screen_info_from_resources (ScreenInfo *info,
     a = g_ptr_array_new ();
     for (i = 0;  i < nmode; ++i)
     {
-	GnomeRRMode *mode;
-
 	g_variant_get_child (modes, i, META_MONITOR_MODE_STRUCT, &id,
-			     NULL, NULL, NULL, NULL);
-	mode = mode_new (info, id);
+			     NULL, NULL, NULL, NULL, NULL);
 
-	g_ptr_array_add (a, mode);
+	g_ptr_array_add (a, mode_new (info, id));
     }
     g_ptr_array_add (a, NULL);
     info->modes = (GnomeRRMode **)g_ptr_array_free (a, FALSE);
@@ -493,17 +488,23 @@ fill_screen_info_from_resources (ScreenInfo *info,
     /* Initialize */
     for (i = 0, crtc = info->crtcs; *crtc; ++i, ++crtc)
     {
-	crtc_initialize (*crtc, g_variant_get_child_value (crtcs, i));
+	GVariant *child = g_variant_get_child_value (crtcs, i);
+	crtc_initialize (*crtc, child);
+	g_variant_unref (child);
     }
 
     for (i = 0, output = info->outputs; *output; ++i, ++output)
     {
-	output_initialize (*output, g_variant_get_child_value (outputs, i));
+	GVariant *child = g_variant_get_child_value (outputs, i);
+	output_initialize (*output, child);
+	g_variant_unref (child);
     }
 
     for (i = 0, mode = info->modes; *mode; ++i, ++mode)
     {
-	mode_initialize (*mode, g_variant_get_child_value (modes, i));
+	GVariant *child = g_variant_get_child_value (modes, i);
+	mode_initialize (*mode, child);
+	g_variant_unref (child);
     }
 
     gather_clone_modes (info);
@@ -537,6 +538,11 @@ fill_out_screen_info (ScreenInfo  *info,
 
     fill_screen_info_from_resources (info, serial, crtcs, outputs,
 				     modes, max_width, max_height);
+
+    g_variant_unref (crtcs);
+    g_variant_unref (outputs);
+    g_variant_unref (modes);
+
     return TRUE;
 }
 
@@ -1119,16 +1125,13 @@ gnome_rr_screen_get_dpms_mode (GnomeRRScreen    *screen,
     g_return_val_if_fail (mode != NULL, FALSE);
 
     power_save = meta_dbus_display_config_get_power_save_mode (screen->priv->proxy);
-
-    if (power_save == META_POWER_SAVE_UNKNOWN) {
+    switch (power_save) {
+    case META_POWER_SAVE_UNKNOWN:
         g_set_error_literal (error,
                              GNOME_RR_ERROR,
                              GNOME_RR_ERROR_NO_DPMS_EXTENSION,
                              "Display is not DPMS capable");
-	return FALSE;
-    }
-
-    switch (power_save) {
+        return FALSE;
     case META_POWER_SAVE_ON:
         *mode = GNOME_RR_DPMS_ON;
         break;
@@ -1164,6 +1167,9 @@ gnome_rr_screen_set_dpms_mode (GnomeRRScreen    *screen,
     g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
     switch (mode) {
+    case GNOME_RR_DPMS_UNKNOWN:
+        power_save = META_POWER_SAVE_UNKNOWN;
+        break;
     case GNOME_RR_DPMS_ON:
         power_save = META_POWER_SAVE_ON;
         break;
@@ -1333,8 +1339,8 @@ output_initialize (GnomeRROutput *output, GVariant *info)
     GPtrArray *a;
     GVariantIter *crtcs, *clones, *modes;
     GVariant *properties, *edid, *tile;
-    int current_crtc_id;
-    guint id;
+    gint32 current_crtc_id;
+    guint32 id;
 
     g_variant_get (info, META_OUTPUT_STRUCT,
 		   &output->id, &output->winsys_id,
@@ -1353,7 +1359,7 @@ output_initialize (GnomeRROutput *output, GVariant *info)
 
 	g_ptr_array_add (a, crtc);
 
-	if (crtc->id == current_crtc_id)
+	if (current_crtc_id != -1 && crtc->id == (guint32) current_crtc_id)
 	{
 	    output->current_crtc = crtc;
 	    append_output_array (&crtc->current_outputs, output);
@@ -1397,6 +1403,7 @@ output_initialize (GnomeRROutput *output, GVariant *info)
     g_variant_lookup (properties, "width-mm", "i", &output->width_mm);
     g_variant_lookup (properties, "height-mm", "i", &output->height_mm);
     g_variant_lookup (properties, "display-name", "s", &output->display_name);
+    g_variant_lookup (properties, "connector-type", "s", &output->connector_type);
     g_variant_lookup (properties, "backlight", "i", &output->backlight);
     g_variant_lookup (properties, "min-backlight-step", "i", &output->min_backlight_step);
     g_variant_lookup (properties, "primary", "b", &output->is_primary);
@@ -1426,6 +1433,8 @@ output_initialize (GnomeRROutput *output, GVariant *info)
 
     if (output->is_primary)
 	output->info->primary = output;
+
+    g_variant_unref (properties);
 }
 
 static GnomeRROutput*
@@ -1441,6 +1450,7 @@ output_copy (const GnomeRROutput *from)
     output->info = from->info;
     output->name = g_strdup (from->name);
     output->display_name = g_strdup (from->display_name);
+    output->connector_type = g_strdup (from->connector_type);
     output->vendor = g_strdup (from->vendor);
     output->product = g_strdup (from->product);
     output->serial = g_strdup (from->serial);
@@ -1488,6 +1498,7 @@ output_free (GnomeRROutput *output)
     g_free (output->product);
     g_free (output->serial);
     g_free (output->display_name);
+    g_free (output->connector_type);
     g_free (output->edid_file);
     if (output->edid)
       g_bytes_unref (output->edid);
@@ -1671,17 +1682,14 @@ gnome_rr_output_get_possible_crtcs (GnomeRROutput *output)
 }
 
 gboolean
-_gnome_rr_output_name_is_builtin_display (const char *name)
+_gnome_rr_output_connector_type_is_builtin_display (const char *connector_type)
 {
-    if (!name)
+    if (!connector_type)
         return FALSE;
 
-    if (strstr (name, "lvds") ||  /* Most drivers use an "LVDS" prefix... */
-	strstr (name, "LVDS") ||
-	strstr (name, "Lvds") ||
-	strstr (name, "LCD")  ||  /* ... but fglrx uses "LCD" in some versions.  Shoot me now, kthxbye. */
-	strstr (name, "eDP")  ||    /* eDP is for internal built-in panel connections */
-	strstr (name, "DSI"))
+    if (strcmp (connector_type, "LVDS") == 0 ||
+	strcmp (connector_type, "eDP") == 0  ||
+	strcmp (connector_type, "DSI") == 0)
         return TRUE;
 
     return FALSE;
@@ -1692,7 +1700,7 @@ gnome_rr_output_is_builtin_display (GnomeRROutput *output)
 {
     g_return_val_if_fail (output != NULL, FALSE);
 
-    return _gnome_rr_output_name_is_builtin_display (output->name);
+    return _gnome_rr_output_connector_type_is_builtin_display (output->connector_type);
 }
 
 /**
@@ -1997,6 +2005,7 @@ crtc_initialize (GnomeRRCrtc *crtc, GVariant *info)
     
     while (g_variant_iter_loop (all_transforms, "u", &transform))
 	crtc->all_transforms |= 1 << transform;
+    g_variant_iter_free (all_transforms);
 }
 
 static void
@@ -2068,15 +2077,22 @@ gnome_rr_mode_get_is_tiled (GnomeRRMode *mode)
     return mode->tiled;
 }
 
+gboolean
+gnome_rr_mode_get_is_interlaced (GnomeRRMode *mode)
+{
+    g_return_val_if_fail (mode != NULL, 0);
+    return (mode->flags & DRM_MODE_FLAG_INTERLACE) != 0;
+}
+
 static void
 mode_initialize (GnomeRRMode *mode, GVariant *info)
 {
     gdouble frequency;
 
-    g_variant_get (info, "(uxuud)",
+    g_variant_get (info, META_MONITOR_MODE_STRUCT,
 		   &mode->id, &mode->winsys_id,
 		   &mode->width, &mode->height,
-		   &frequency);
+		   &frequency, &mode->flags);
     
     mode->freq = frequency * 1000;
 }
@@ -2229,6 +2245,14 @@ gnome_rr_output_supports_underscanning (GnomeRROutput *output)
     return output->supports_underscanning;
 }
 
+const char *
+_gnome_rr_output_get_connector_type (GnomeRROutput *output)
+{
+    g_return_val_if_fail (output != NULL, NULL);
+
+    return output->connector_type;
+}
+
 gboolean
 _gnome_rr_output_get_tile_info (GnomeRROutput *output,
 				GnomeRRTile *tile)
